"""Functions for analysis."""
import re

import pandas as pd
import numpy as np


def compute_crossfeed(data_frame: pd.DataFrame, tol=1e-4):
    """Compute crossfeed."""
    return data_frame.apply(_find_cross_feed, axis=1, tol=tol)


def compute_directed_crossfeed(data_frame: pd.DataFrame, tol=1e-4):
    """Compute directed crossfeed."""
    return data_frame.apply(_find_directed_cross_feed, axis=1, tol=tol)


def get_suppliers(data_frame: pd.DataFrame):
    """Get suppliers.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """

    df = data_frame.reset_index()
    selector = ["strain"] + [column for column in df.columns if column.startswith("y_")]
    return df[selector].apply(
        lambda row: frozenset(row.index[row == 1]) - frozenset([f"y_{row.strain}"]),
        axis=1,
    )


def get_communities(data_frame: pd.DataFrame):
    """Get communities.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    selector = ["strain"] + [column for column in df.columns if column.startswith("y_")]
    return df[selector].apply(
        lambda row: (frozenset(row.index[row == 1])),
        axis=1,
    )


def find_focal_strain_growth(data_frame: pd.DataFrame):
    """Find focal strain growth and add to DataFrame.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    data_frame["strain_growth"] = data_frame.apply(
        lambda row: row[f"Growth_{row.name[1]}"], axis=1
    )
    return data_frame


def get_unique_strains(data_frame: pd.DataFrame):
    """Get all strains included in the dataset.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    regex = re.compile(r"Growth_(\w+)")
    return [regex.match(c).groups()[0] for c in data_frame.columns if regex.match(c)]


def get_unique_focal_strains(data_frame: pd.DataFrame):
    """Get all focal strains included in the dataset.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    return data_frame.reset_index().strain.unique()


def get_unique_carbon_sources(data_frame: pd.DataFrame):
    """Get all unique carbon sources in the dataset.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    return data_frame.reset_index().carbon_source.unique()


def count_carbon_sources_for_strain(data_frame: pd.DataFrame):
    """Count the number of viable environments for each strain.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    strains = get_unique_strains(df)
    unique_carbon_sources = [
        len(df[df[f"Growth_{strain}"] > 0].carbon_source.unique()) for strain in strains
    ]
    return pd.DataFrame({"strain": strains, "carbon_sources": unique_carbon_sources})


def count_carbon_sources_as_focal_strain(data_frame: pd.DataFrame):
    """Count the number of viable environments as focal strain.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    strains = get_unique_strains(df)
    unique_carbon_sources = [
        len(df[(df.strain == strain) & (df.growth_rate > 0)].carbon_source.unique())
        for strain in strains
    ]
    return pd.DataFrame(
        {"strain": strains, "focal_carbon_sources": unique_carbon_sources}
    )


def count_carbon_sources_for_isolation(data_frame: pd.DataFrame):
    """Count the number of viable environments as focal strain.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    strains = get_unique_strains(df)
    supplied = np.array(map(lambda x: len(x) == 0, get_suppliers(df)))
    unique_carbon_sources = [
        len(
            df[
                (df.strain == strain) & (df.growth_rate > 0) & supplied
            ].carbon_source.unique()
        )
        for strain in strains
    ]
    return pd.DataFrame(
        {"strain": strains, "isolated_carbon_sources": unique_carbon_sources}
    )


def count_carbon_sources_as_supplier_strain(data_frame: pd.DataFrame):
    """Count the number of viable environments as supplier strain.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    strains = get_unique_strains(df)
    unique_carbon_sources = [
        len(
            df[
                (df.strain != strain) & (df[f"Growth_{strain}"] > 0)
            ].carbon_source.unique()
        )
        for strain in strains
    ]
    return pd.DataFrame(
        {"strain": strains, "supplier_carbon_sources": unique_carbon_sources}
    )


def count_communities_as_supplier_strain(data_frame: pd.DataFrame):
    """Count the number of communities as supplier strain.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    strains = get_unique_strains(df)
    unique_communities = [
        len(df[(df.strain != strain) & (df[f"Growth_{strain}"] > 0)])
        for strain in strains
    ]
    return pd.DataFrame({"strain": strains, "communities": unique_communities})


def count_isolated_strains_on_carbon_source(data_frame: pd.DataFrame):
    """Count the number of strains in isolation on a carbon source.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    indicator_columns = [column for column in df.columns if column.startswith("y_")]
    return pd.DataFrame(
        df[df[indicator_columns].sum(axis=1) == 1]
        .groupby("carbon_source")
        .size()
        .rename("isolated_strains")
    )


def count_non_redundant_supplying_communities_on_carbon_source(
    data_frame: pd.DataFrame,
):
    """Count the number of non-redundant supplying communities on a carbon source.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    supplying_communities = pd.DataFrame(
        {
            "carbon_source": df.carbon_source,
            "supplying_community": get_suppliers(df),
        }
    )
    return (
        supplying_communities.groupby("carbon_source")
        .apply(lambda group: len(frozenset(group.supplying_community)) - 1)
        .rename("supplying_community")
        .to_frame()
    )


def count_exchanged_metabolites_on_carbon_source(data_frame: pd.DataFrame):
    """Count the number of metabolites exchanged on a carbon source.

    Parameters
    ----------
    data_frame : pandas.DataFrame
        A solution dataframe from misosoup.
    """
    df = data_frame.reset_index()
    metabolites_exchanged = pd.DataFrame(
        {
            "carbon_source": df.carbon_source,
            "metabolites_exchanged": compute_crossfeed(df),
        }
    )
    return (
        metabolites_exchanged.groupby("carbon_source")
        .apply(lambda group: len(frozenset(group.metabolites_exchanged)) - 1)
        .rename("metabolites_exchanged")
        .to_frame()
    )


def _find_cross_feed(row, tol=1e-4):
    positive = set()
    negative = set()
    for rid, val in row.items():
        compound = re.search("(?<=R_EX_).*(?=_e)", rid)
        if compound and rid.endswith("_i"):
            global_rid = f"R_EX_{compound}_e"
            global_val = row[global_rid] if global_rid in row else 0
            adjusted_val = val - global_val
            if adjusted_val < -tol:
                negative.add(compound.group(0))
            elif adjusted_val > tol:
                positive.add(compound.group(0))
    return frozenset(positive & negative)


def _find_directed_cross_feed(row, tol=1e-4):
    strain = row.name[1]
    crossfeed = _find_cross_feed(row, tol)
    directed_crossfeed = {}
    for compound in crossfeed:
        rid = f"R_EX_{compound}_e_{strain}_i"
        directed_crossfeed[compound] = row[rid] if rid in row else 0
    return frozenset(directed_crossfeed)



"""MiSoS(oup)"""

import argparse
import glob
import logging
import os
from collections import defaultdict

import yaml

from .library.getters import get_biomass, get_exchange_reactions
from .library.readwrite import load_models, read_compounds
from .library.validate import validate_solution_dict
from .library.minimizer import Minimizer

from .reframed.layered_community import LayeredCommunity


def main(args):
    """Main function."""
    input_paths = glob.glob(args.input[0]) if len(args.input) == 1 else args.input
    models = load_models(input_paths)

    media = read_compounds(args.media)
    base_medium = media["base_medium"] if "base_medium" in media.keys() else {}

    community = LayeredCommunity("CoI", models, copy_models=False)

    solutions = defaultdict(dict)

    if args.objective:
        objective = {reaction: 1 for reaction in args.objective}
    else:
        objective = {community.merged_model.biomass_reaction: 1}

    if args.disable_objective:
        objective = None

    for medium_id, medium_composition in media.items():
        if not medium_id == "base_medium" and (
            not args.media_select or medium_id in args.media_select
        ):
            medium = {
                **medium_composition,
                **base_medium,
            }

            minimizer = Minimizer(
                org_id=args.strain,
                medium=medium,
                community=community,
                values=(
                    get_biomass(community)
                    + get_exchange_reactions(community.merged_model)
                ),
                community_size=args.community_size,
                objective=objective,
                parsimony=args.parsimony,
                minimal_growth=args.minimal_growth,
                cache_file=args.cache_file,
            )

            solution = minimizer.minimize()

            solutions[medium_id][args.strain] = solution

    output_dict = {
        k: {org: sol if sol else [{f"Growth_{org}": 0}] for org, sol in v.items()}
        for k, v in solutions.items()
    }

    if args.validate:
        validate_solution_dict(output_dict, args.exchange_format)

    output = yaml.dump(output_dict)

    if args.output:
        if not os.path.exists(os.path.dirname(args.output)):
            os.makedirs(os.path.dirname(args.output))
        with open(args.output, "a", encoding="utf8") as file_descriptor:
            file_descriptor.write(output)
    else:
        print(output)


def entry():
    """Misosoup entry point."""
    parser = argparse.ArgumentParser(
        description="""
    Compute minimal supplying communities with MiSoS(oup).

    `misosoup` will create a large community network and evaluate which
    communities can supply growth of some strain (or each other, if no strain
    is chosen).

    It will report the results in `yaml` format, where the dictionaries contain
    all active exchange reactions and their rates, as well as the activity of
    strains (i.e. `y_strain: 1` for any active strain), the community growth
    rate and the individual growth rates of all participating strains:

        ```
        carbon_source:
            strain:
                - {}
                - {}
        ```
    """,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "input", nargs="+", type=str, help="List or wildcard for model paths."
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Output file. Format: YAML. If not supplied, will print to stdout.",
    )
    parser.add_argument(
        "--cache-file",
        type=str,
        help="Path to cache file. If set, intermediate constraints will be stored.",
    )
    parser.add_argument(
        "--media",
        type=str,
        required=True,
        help="Path to media. Format: YAML. File needs to contain dictionary with media, see examples.",
    )
    parser.add_argument(
        "--media-select",
        type=str,
        nargs="+",
        help="List of media names to use from the media file.",
    )
    parser.add_argument(
        "--strain",
        type=str,
        default="min",
        help="Focal strain model id. If no strain is provided, we compute minimal communities.",
    )
    parser.add_argument(
        "--parsimony", action="store_true", help="Compute parsimony solution."
    )
    parser.add_argument(
        "--community-size",
        type=int,
        default=0,
        help="Maximum community size. Default: 0 (Arbitrary).",
    )
    parser.add_argument(
        "--minimal-growth",
        type=float,
        default=0.01,
        help=(
            "Minimal required growth for strain or community."
            "Each strain that is considered to grow needs to at least achieve "
            "this minimal growth rate given by this argument."
        ),
    )
    parser.add_argument(
        "--exchange-format",
        type=str,
        default="R_EX_{}_e",
        help=(
            "Regular expression to retrieve the carbon source from an exchange "
            "reaction. The group containing the carbon source, should be named "
            "`carbon_source`; as can be seen from the default. "
            "default: `R_EX_(\\w+)_e"
        ),
    )
    parser.add_argument(
        "--disable-objective",
        action="store_true",
        help="If set, no objective will be optimized.",
    )
    parser.add_argument(
        "--objective",
        type=str,
        nargs="+",
        help=(
            "List of ids that are part of the objective function. By default "
            + "the community biomass is maximized."
        ),
    )
    parser.add_argument(
        "--validate",
        action="store_true",
        help=(
            "Validate solution. Numerical consistency will be verified on the "
            "final solution."
        ),
    )
    parser.add_argument("--log", default="INFO", help="Log level. Default: INFO")

    args_parsed = parser.parse_args()

    logging.basicConfig(level=args_parsed.log, format="%(asctime)s %(message)s")

    main(args_parsed)



"""Read and write utility functions."""
import yaml

import pandas as pd

from reframed.io.sbml import load_cbmodel

from .common import get_reaction_name


def load_models(paths):
    return [load_cbmodel(path, flavor="fbc2") for path in paths]


def read_medium(path, medium_name):
    """Read medium."""
    with open(path, "r", encoding="utf8") as file_descriptor:
        medium_data = yaml.safe_load(file_descriptor)
    return {
        key: [get_reaction_name(compound) for compound in compounds]
        for key, compounds in medium_data[medium_name].items()
    }


def read_compounds(path):
    """Read compounds."""
    with open(path, "r", encoding="utf8") as file_descriptor:
        compounds = yaml.safe_load(file_descriptor)
    return compounds


def read_supplements(path):
    """Read supplements."""
    with open(path, "r", encoding="utf8") as file_descriptor:
        supplement_sources = yaml.safe_load(file_descriptor)
    return [get_reaction_name(source) for source in supplement_sources.carbon_sources]


def write_minimal_suppliers(solutions, path=None):
    """Write minimal suppliers."""
    processed = {
        org: [
            {
                "biomass": sol.values["community_growth"],
                "community": [
                    k for k, v in sol.values.items() if v == 1 and k.startswith("y_")
                ],
            }
            for sol in sols
        ]
        for org, sols in solutions.items()
    }
    if path:
        with open(path, "w", encoding="utf8") as file_descriptor:
            file_descriptor.write(yaml.dump(processed))
    else:
        print(yaml.dump(processed))


def read_solutions_yaml(file_path):
    """Read solutions yaml."""
    with open(file_path, "r", encoding="utf8") as file_descriptor:
        solutions = yaml.safe_load(file_descriptor)

    carbon_source_solutions_type = list(solutions.values())[0]
    strain_solutions_type = list(carbon_source_solutions_type.values())[0]
    data_dict = (
        {
            (carbon_source, strain, idx): {**strain_solution}
            for carbon_source, carbon_source_solutions in solutions.items()
            for strain, strain_solutions in carbon_source_solutions.items()
            for idx, strain_solution in enumerate(strain_solutions)
        }
        if isinstance(strain_solutions_type, list)
        else {
            (carbon_source, strain, 0): {**strain_solutions}
            for carbon_source, carbon_source_solutions in solutions.items()
            for strain, strain_solutions in carbon_source_solutions.items()
        }
    )

    data = pd.DataFrame.from_dict(
        data_dict,
        orient="index",
    ).sort_index(level=0)
    data.index.names = ["carbon_source", "strain", "solution_idx"]
    data["growth_rate"] = data.community_growth

    return data


def read_isolates_yaml(file_path):
    """Read isolates yaml."""
    with open(file_path, "r", encoding="utf8") as file_descriptor:
        solutions = yaml.safe_load(file_descriptor)

    # carbon_sources = list(solutions.keys())
    # strains = list(solutions[carbon_sources[0]].keys())

    isolates_df = pd.DataFrame.from_dict(
        {
            (carbon_source, strain): {
                "biomass": strain_solution["biomass"],
                **strain_solution["exchange"],
            }
            for carbon_source, carbon_source_solutions in solutions.items()
            for strain, strain_solution in carbon_source_solutions.items()
        },
        orient="index",
    )
    isolates_df.index.names = ["carbon_source", "strain"]
    isolates_df["growth_rate"] = isolates_df.biomass

    return isolates_df



"""Minimizer module."""

import logging
import math
import os
import yaml

from reframed.solvers import GurobiSolver
from reframed.solvers.solver import VarType
from reframed.solvers.solution import Status

from ..reframed.layered_community import LayeredCommunity

BOUND_INF = 1000


class Minimizer:
    """Minimizer class."""

    def __init__(
        self,
        org_id: str,
        medium: dict,
        community: LayeredCommunity,
        values: list,
        community_size: int,
        objective: dict,
        parsimony: bool,
        minimal_growth: float,
        cache_file: str = None,
    ):
        """Initialize `Minimize`."""
        self.community = community
        self.medium = medium
        self.minimal_growth = minimal_growth
        self.values = values
        self.community_size = community_size
        self.objective = objective
        self.parsimony = parsimony

        self.solver = GurobiSolver(community.merged_model)

        self._setup_binary_variables()
        self._setup_medium()

        if parsimony:
            self._setup_parsimony()

        if not org_id or org_id == "min":
            self.solver.add_constraint(
                "c_community_growth",
                {community.merged_model.biomass_reaction: 1},
                ">",
                minimal_growth,
                update=True,
            )
        else:
            self.solver.add_constraint(
                f"c_{org_id}_focal",
                {f"y_{org_id}": 1},
                ">",
                1,
                update=True,
            )

        self._community_objective = {
            f"y_{org_id}": 1 for org_id in self.community.organisms.keys()
        }

        self.solutions = []
        self.community_constraints = {}
        self.knowledge_constraints = {}

        self.cache_file = cache_file
        if cache_file and os.path.exists(cache_file):
            with open(cache_file, encoding="utf8") as cache_fd:
                cache = yaml.safe_load(cache_fd)
            self._load_constraints_from_cache(cache)

    def minimize(self):
        """Minimize community."""
        i = len(self.community_constraints)
        while True:
            if self.cache_file:
                self._dump_constraints_to_cache()

            logging.info("------------")
            logging.info("Starting optimization...")

            solution = self._minimize_community()

            if solution.status != Status.OPTIMAL:
                logging.info("Solution status: %s", str(solution.status))
                break

            selected = {
                k: 1
                for k in self._community_objective.keys()
                if solution.values[k] > 0.5
            }
            not_selected = {
                k: 1
                for k in self._community_objective.keys()
                if solution.values[k] < 0.5
            }

            logging.info("Community size: %i", len(selected))
            logging.info(
                "Community growth: %f",
                solution.values[self.community.merged_model.biomass_reaction],
            )

            # Fix selected community
            self.solver.add_constraint(
                "c_not_selection", not_selected, "=", 0, update=True
            )
            self.solver.add_constraint(
                "c_selection", selected, "=", len(selected), update=True
            )

            try:
                solution = self._no_objective_optimization()

                if not self._check_solution(solution, selected, not_selected):
                    logging.info(
                        "Community Inconsistent: %s", str(list(selected.keys()))
                    )
                    self._add_knowledge_constraint(not_selected)
                    continue

                growth = 0
                if self.objective:
                    objective_solution = self._objective_optimization()
                    if not self._check_solution(
                        objective_solution, selected, not_selected
                    ):
                        logging.warning("Unable to optimize objective.")
                        growth = -1
                    else:
                        growth = objective_solution.values[
                            self.community.merged_model.biomass_reaction
                        ]
                        solution = objective_solution

                if self.parsimony:
                    parsimony_solution = self._parsimony_optimization(growth)
                    if not self._check_solution(
                        parsimony_solution, selected, not_selected
                    ):
                        logging.warning("Unable to minimize fluxes.")
                    else:
                        solution = parsimony_solution

            finally:
                self.solver.remove_constraints(["c_not_selection", "c_selection"])

            self.solver.add_constraint(
                f"c_{i}", selected, "<", len(selected) - 1, update=True
            )
            self.community_constraints[f"c_{i}"] = selected

            if self.community_size and len(selected) > self.community_size:
                break

            self.solutions.append(
                {var: rate for var, rate in solution.values.items() if rate}
            )

            i += 1

        if self.cache_file:
            self._dump_constraints_to_cache()

        return self.solutions

    @property
    def _get_values(self):
        return list(self._community_objective.keys()) + self.values

    def _minimize_community(self):
        return self.solver.solve(
            linear=self._community_objective,
            get_values=self._get_values,
            minimize=True,
        )

    def _objective_optimization(self):
        logging.info("Starting growth optimization.")
        return self.solver.solve(
            linear=self.objective,
            get_values=self._get_values,
            minimize=False,
        )

    def _no_objective_optimization(self):
        logging.info("Starting no optimization.")
        return self.solver.solve(
            get_values=self._get_values,
        )

    def _parsimony_optimization(self, growth: float = 0):
        logging.info("Starting parsimony optimization.")
        if growth > 0:
            self.solver.add_constraint(
                "c_growth",
                self.objective,
                ">",
                growth,
                update=True,
            )

        solution = self.solver.solve(
            linear={
                f"abs_{rid}_{sense}": 1
                for rid in self.community.merged_model.reactions
                for sense in ["pos", "neg"]
            },
            get_values=self._get_values,
            minimize=True,
        )

        self.solver.remove_constraints(["c_growth"])

        return solution

    def _check_solution(self, solution, selected, not_selected):
        logging.debug("Solution status: %s", str(solution.status))
        if solution.status != Status.OPTIMAL:
            return False
        logging.info(
            "Community growth: %f",
            solution.values[self.community.merged_model.biomass_reaction],
        )
        return True

    def _add_knowledge_constraint(self, not_selected: dict):
        constraint_name = f"c_tmp_{len(self.knowledge_constraints) + 1}"
        self.solver.add_constraint(constraint_name, not_selected, ">", 1)
        self.knowledge_constraints[constraint_name] = not_selected

    def _load_constraints_from_cache(self, cache: dict):
        for name, selected in cache["community_constraints"].items():
            self.solver.add_constraint(name, selected, "<", len(selected) - 1)
            self.community_constraints[name] = selected

        for name, not_selected in cache["knowledge_constraints"].items():
            self.solver.add_constraint(name, not_selected, ">", 1)
            self.knowledge_constraints[name] = not_selected

        self.solutions = cache["solutions"]

        self.solver.update()

    def _dump_constraints_to_cache(self):
        with open(self.cache_file, "w", encoding="utf8") as cache_fd:
            yaml.dump(
                {
                    "community_constraints": self.community_constraints,
                    "knowledge_constraints": self.knowledge_constraints,
                    "solutions": self.solutions,
                },
                cache_fd,
            )

    def _setup_parsimony(self):
        # add absolute variables for each reaction
        for rid in self.community.merged_model.reactions:
            self.solver.add_variable(f"abs_{rid}_pos", 0, 1000, update=False)
            self.solver.add_variable(f"abs_{rid}_neg", 0, 1000, update=False)

        self.solver.update()

        # add absolute constraints for each reaction
        for rid in self.community.merged_model.reactions:
            self.solver.add_constraint(
                f"c_{rid}_abs",
                {f"abs_{rid}_pos": 1, f"abs_{rid}_neg": -1},
                "=",
                0,
            )

    def _setup_binary_variables(self):
        """Add binary variables to the solver instance."""
        for org_id in self.community.organisms.keys():
            self.solver.add_variable(
                f"y_{org_id}",
                0,
                1,
                vartype=VarType.BINARY,
                update=False,
            )

        self.solver.update()

        for org_id, org_model in self.community.organisms.items():
            org_var = f"y_{org_id}"
            for r_id, reaction in org_model.reactions.items():
                if (
                    not r_id.startswith("R_EX")
                    and r_id != org_model.biomass_reaction
                    and reaction.lb * reaction.ub <= 0
                ):
                    continue

                merged_id = self.community.reaction_map[(org_id, r_id)]
                ubound = BOUND_INF
                lbound = -BOUND_INF

                if r_id == org_model.biomass_reaction:
                    lbound = self.minimal_growth

                if reaction.lb * reaction.ub > 0:
                    lbound = -BOUND_INF if math.isinf(reaction.lb) else reaction.lb
                    ubound = BOUND_INF if math.isinf(reaction.ub) else reaction.ub
                    self.solver.set_bounds({merged_id: (-BOUND_INF, BOUND_INF)})

                self.solver.add_constraint(
                    f"c_{merged_id}_lb",
                    {merged_id: 1, org_var: -lbound},
                    ">",
                    0,
                    update=False,
                )
                self.solver.add_constraint(
                    f"c_{merged_id}_ub",
                    {merged_id: 1, org_var: -ubound},
                    "<",
                    0,
                    update=False,
                )

    def _setup_medium(self):
        """Setup the medium for model on solver."""
        missing_reactions = set(self.medium.keys()) - set(
            self.community.merged_model.reactions.keys()
        )
        for r_id in missing_reactions:
            logging.warning(
                "Missing reaction %s in model %s", r_id, self.community.merged_model.id
            )
        for r_id in self.community.merged_model.reactions.keys():
            if r_id.startswith("R_EX_") and not r_id.endswith("_i"):
                bound = self.medium[r_id] if r_id in self.medium.keys() else 0
                self.solver.add_constraint(
                    f"c_{r_id}_lb",
                    {r_id: 1},
                    ">",
                    bound,
                )



"""Validate."""
import re
import logging
import yaml


def validate_solution(solution, exchange_format, exclude_strains=None):
    """Validate solution."""
    valid = True
    if exclude_strains is None:
        exclude_strains = []
    active_strains = []
    active_exchanges = []

    for var, rate in solution.items():
        if rate:
            if var.startswith("y_"):
                active_strains.append(var[2:])
            else:
                active_exchanges.append(var)

    matchers = [
        re.compile(exchange_format.format(r"\w+") + f"_{strain}_i")
        for strain in active_strains
        if strain not in exclude_strains
    ]

    for var in active_exchanges:
        if var.endswith("_i") and not any(m.match(var) for m in matchers):
            logging.warning("Numerical inconsistency found for %s", var)
            valid = False

    return valid


def validate_solution_dict(solution_data, exchange_format):
    """Validate if a solution dictionary reports numerical inconsistency."""
    valid = True
    for carbon_source, carbon_source_solutions in solution_data.items():
        for _, org_sols in carbon_source_solutions.items():
            for _, sol in enumerate(org_sols):
                if not validate_solution(sol, exchange_format):
                    valid = False
                    print(
                        f"Inconsistency found: carbon source={carbon_source}, "
                        "organism={org}, solution={i}"
                    )
    return valid


def validate_solution_file(path, exchange_format="R_EX_{}_e"):
    """Validate if a solution file reports numerical inconsistency."""
    with open(path, encoding="utf8") as file_descriptor:
        solution_data = yaml.safe_load(file_descriptor)
    return validate_solution_dict(solution_data, exchange_format)



"""Gurobi solver instance for individual environments."""

from reframed.solvers.solver import Solver
from reframed.solvers.gurobi_solver import GurobiSolver, default_parameters
from gurobipy import Model as GurobiModel


class GurobiEnvSolver(GurobiSolver):
    """Gurobi interface initialized in gurobi environment."""

    def __init__(self, model=None, env=None):
        """Init GurobiEnvSolver."""
        Solver.__init__(self)
        if env:
            self.problem = GurobiModel(env=env)
        else:
            self.problem = GurobiModel()
        self.set_parameters(default_parameters)
        if model:
            self.build_problem(model)
